{ lib, name, ... }:

let
  inherit (lib) mkOption;

  inherit (lib.types)
    attrsOf
    bool
    ints
    listOf
    nullOr
    str
    submodule
    ;
in

{
  options = {
    deployment = {
      systemType = mkOption {
        description = ''
          System type used for this node, e.g. NixOS.
        '';
        default = "nixos";
        # TODO: enum among all registered systems?
        type = str;
      };
      targetHost = mkOption {
        description = ''
          The target SSH node for deployment.

          By default, the node's attribute name will be used.
          If set to null, only local deployment will be supported.
        '';
        type = nullOr str;
        default = name;
      };
      targetPort = mkOption {
        description = ''
          The target SSH port for deployment.

          By default, the port is the standard port (22) or taken
          from your ssh_config.
        '';
        type = nullOr ints.unsigned;
        default = null;
      };
      targetUser = mkOption {
        description = ''
          The user to use to log into the remote node. If set to null, the
          target user will not be specified in SSH invocations.
        '';
        type = nullOr str;
        default = "root";
      };
      allowLocalDeployment = mkOption {
        description = ''
          Allow the configuration to be applied locally on the host running
          Colmena.

          For local deployment to work, all of the following must be true:
          - The node must be running NixOS.
          - The node must have deployment.allowLocalDeployment set to true.
          - The node's networking.hostName must match the hostname.

          To apply the configurations locally, run `colmena apply-local`.
          You can also set deployment.targetHost to null if the nost is not
          accessible over SSH (only local deployment will be possible).
        '';
        type = bool;
        default = false;
      };
      buildOnTarget = mkOption {
        description = ''
          Whether to build the system profiles on the target node itself.

          When enabled, Colmena will copy the derivation to the target
          node and initiate the build there. This avoids copying back the
          build results involved with the native distributed build
          feature. Furthermore, the `build` goal will be equivalent to
          the `push` goal. Since builds happen on the target node, the
          results are automatically "pushed" and won't exist in the local
          Nix store.

          You can temporarily override per-node settings by passing
          `--build-on-target` (enable for all nodes) or
          `--no-build-on-target` (disable for all nodes) on the command
          line.
        '';
        type = bool;
        default = false;
      };
      tags = mkOption {
        description = ''
          A list of tags for the node.

          Can be used to select a group of nodes for deployment.
        '';
        type = listOf str;
        default = [ ];
      };
      keys = mkOption {
        description = ''
          A set of secrets to be deployed to the node.

          Secrets are transferred to the node out-of-band and
          never ends up in the Nix store.
        '';
        type = attrsOf (submodule (import ./key.nix));
        default = { };
      };
      replaceUnknownProfiles = mkOption {
        description = ''
          Allow a configuration to be applied to a host running a profile we
          have no knowledge of. By setting this option to false, you reduce
          the likelyhood of rolling back changes made via another Colmena user.

          Unknown profiles are usually the result of either:
          - The node had a profile applied, locally or by another Colmena.
          - The host running Colmena garbage-collecting the profile.

          To force profile replacement on all targeted nodes during apply,
          use the flag `--force-replace-unknown-profiles`.
        '';
        type = bool;
        default = true;
      };
      privilegeEscalationCommand = mkOption {
        description = ''
          Command to use to elevate privileges when activating the new profiles on SSH hosts.

          This is used on SSH hosts when `deployment.targetUser` is not `root`.
          The user must be allowed to use the command non-interactively.
        '';
        type = listOf str;
        default = [
          "sudo"
          "-H"
          "--"
        ];
      };
      sshOptions = mkOption {
        description = ''
          Extra SSH options to pass to the SSH command.
        '';
        type = listOf str;
        default = [ ];
      };
    };
  };
}
