//! Node filters.

use std::collections::HashSet;
use std::convert::AsRef;
use std::iter::Iterator;
use std::str::FromStr;

use clap::Args;
use glob::Pattern as GlobPattern;

use super::{ColmenaError, ColmenaResult, NodeConfig, NodeName};

#[derive(Debug, Default, Args)]
pub struct NodeFilterOpts {
    #[arg(
        long,
        value_name = "NODES",
        help = "Node selector",
        long_help = r#"Select a list of nodes to deploy to.

The list is comma-separated and globs are supported. To match tags, prepend the filter by @. Valid examples:

- host1,host2,host3
- edge-*
- edge-*,core-*
- @a-tag,@tags-can-have-*"#
    )]
    pub on: Option<NodeFilter>,
}

/// A filter rule.
#[derive(Debug, Clone, Eq, PartialEq)]
pub enum NodeFilter {
    /// Matches a node's attribute name.
    MatchName(GlobPattern),

    /// Matches a node's `deployment.tags`.
    MatchTag(GlobPattern),

    /// Matches an Union
    Union(Vec<Box<NodeFilter>>),

    /// Matches an Intersection
    Inter(Vec<Box<NodeFilter>>),

    /// Matches the complementary
    Not(Box<NodeFilter>),

    /// Empty
    Empty,
}

impl FromStr for NodeFilter {
    type Err = ColmenaError;
    fn from_str(s: &str) -> Result<Self, Self::Err> {
        Self::new(s)
    }
}

#[inline]
fn end_delimiter(c: char) -> bool {
    [',', '&', ')'].contains(&c)
}

impl NodeFilter {
    fn and(a: Self, b: Self) -> Self {
        match (a, b) {
            (Self::Inter(mut av), Self::Inter(mut bv)) => {
                av.append(&mut bv);
                Self::Inter(av)
            }
            (Self::Inter(mut av), b) => {
                av.push(Box::new(b));
                Self::Inter(av)
            }
            (a, Self::Inter(mut bv)) => {
                bv.push(Box::new(a));
                Self::Inter(bv)
            }
            (a, b) => Self::Inter(vec![Box::new(a), Box::new(b)]),
        }
    }

    fn or(a: Self, b: Self) -> Self {
        match (a, b) {
            (Self::Union(mut av), Self::Union(mut bv)) => {
                av.append(&mut bv);
                Self::Union(av)
            }
            (Self::Union(mut av), b) => {
                av.push(Box::new(b));
                Self::Union(av)
            }
            (a, Self::Union(mut bv)) => {
                bv.push(Box::new(a));
                Self::Union(bv)
            }
            (a, b) => Self::Union(vec![Box::new(a), Box::new(b)]),
        }
    }

    fn not(a: Self) -> Self {
        if let Self::Not(ae) = a {
            *ae
        } else {
            Self::Not(Box::new(a))
        }
    }

    /// Parses an elementary expression,
    /// that is base tags and name, with expression between parentheses
    /// Negations are also parsed here as the most prioritary operation
    ///
    /// It returns the unparsed text that follows
    fn parse_expr0(unparsed: &str) -> ColmenaResult<(Self, &str)> {
        let unparsed = unparsed.trim_start();
        // Negation
        if let Some(negated_expr) = unparsed.strip_prefix('!') {
            let (negated, unparsed) = Self::parse_expr0(negated_expr)?;
            Ok((Self::not(negated), unparsed))
        } else
        // parentheses
        if let Some(parenthesed_expr) = unparsed.strip_prefix('(') {
            let (interior, unparsed) = Self::parse_expr2(parenthesed_expr)?;
            Ok((
                interior,
                unparsed.strip_prefix(')').ok_or(ColmenaError::Unknown {
                    message: format!("Expected a closing parenthesis at {:?}.", unparsed),
                })?,
            ))
        } else
        // tag
        if let Some(tag_expr) = unparsed.strip_prefix('@') {
            match tag_expr
                .find(end_delimiter)
                .map(|idx| tag_expr.split_at(idx))
                .map(|(tag, end)| (tag.trim_end(), end))
            {
                Some((tag, unparsed)) => {
                    if tag.is_empty() {
                        return Err(ColmenaError::EmptyFilterRule);
                    } else {
                        Ok((Self::MatchTag(GlobPattern::new(tag).unwrap()), unparsed))
                    }
                }
                None => {
                    let tag_expr = tag_expr.trim_end();
                    if tag_expr.is_empty() {
                        Err(ColmenaError::EmptyFilterRule)
                    } else {
                        Ok((Self::MatchTag(GlobPattern::new(tag_expr).unwrap()), ""))
                    }
                }
            }
        } else
        //node name
        {
            match unparsed
                .find(end_delimiter)
                .map(|idx| unparsed.split_at(idx))
                .map(|(tag, end)| (tag.trim_end(), end))
            {
                Some((name, unparsed)) => {
                    if name.is_empty() {
                        Err(ColmenaError::EmptyFilterRule)
                    } else {
                        Ok((Self::MatchName(GlobPattern::new(name).unwrap()), unparsed))
                    }
                }
                None => {
                    let unparsed = unparsed.trim_end();
                    if unparsed.is_empty() {
                        Err(ColmenaError::EmptyFilterRule)
                    } else {
                        Ok((Self::MatchName(GlobPattern::new(unparsed).unwrap()), ""))
                    }
                }
            }
        }
    }

    /// Parses the union operations between elementary expression.
    ///
    /// It returns the unparsed text that follows
    fn parse_op1(acc: Self, unparsed: &str) -> ColmenaResult<(Self, &str)> {
        let unparsed = unparsed.trim_start();
        if let Some(unions) = unparsed.strip_prefix(',') {
            let (base_expr, unparsed) = Self::parse_expr0(unions)?;
            Self::parse_op1(Self::or(acc, base_expr), unparsed)
        } else {
            Ok((acc, unparsed))
        }
    }

    /// Parses elementary expression and their unions.
    ///
    /// It returns the unparsed text that follows
    fn parse_expr1(unparsed: &str) -> ColmenaResult<(Self, &str)> {
        let (base_expr, unparsed) = Self::parse_expr0(unparsed)?;
        Self::parse_op1(base_expr, unparsed)
    }

    /// Parses the intersection operations between unions.
    ///
    /// It returns the unparsed text that follows
    fn parse_op2(acc: Self, unparsed: &str) -> ColmenaResult<(Self, &str)> {
        if let Some(intersections) = unparsed.strip_prefix('&') {
            let (union, unparsed) = Self::parse_expr1(intersections)?;
            Self::parse_op2(Self::and(acc, union), unparsed)
        } else {
            Ok((acc, unparsed))
        }
    }

    /// Parses a complete expression
    ///
    /// It returns the unparsed text that follows
    fn parse_expr2(unparsed: &str) -> ColmenaResult<(Self, &str)> {
        let (union, unparsed) = Self::parse_expr1(unparsed)?;
        Self::parse_op2(union, unparsed)
    }

    /// Creates a new filter using an expression passed using `--on`.
    pub fn new<S: AsRef<str>>(filter: S) -> ColmenaResult<Self> {
        let filter = filter.as_ref();
        let trimmed = filter.trim();

        if trimmed.is_empty() {
            log::warn!("Filter \"{}\" is blank and will match nothing", filter);

            return Ok(Self::Empty);
        }
        let (target_filter, unparsed) = Self::parse_expr2(trimmed)?;
        if unparsed != "" {
            Err(ColmenaError::Unknown {
                message: format!("Found garbage {:?} when parsing the node filter.", unparsed),
            })
        } else {
            Ok(target_filter)
        }
    }

    /// Returns whether the filter has any rule matching NodeConfig information.
    ///
    /// Evaluating `config.deployment` can potentially be very expensive,
    /// especially when its values (e.g., tags) depend on other parts of
    /// the configuration.
    pub fn has_node_config_rules(&self) -> bool {
        match self {
            Self::MatchName(_) => false,
            Self::MatchTag(_) => true,
            Self::Union(v) => v.iter().any(|e| e.has_node_config_rules()),
            Self::Inter(v) => v.iter().any(|e| e.has_node_config_rules()),
            Self::Not(e) => e.has_node_config_rules(),
            Self::Empty => false,
        }
    }

    /// Decides whether a node is accepted by the filter or not.
    /// panic if the filter depends on tags and config is None
    fn is_accepted(&self, name: &NodeName, config: Option<&NodeConfig>) -> bool {
        match self {
            Self::MatchName(pat) => pat.matches(name.as_str()),
            Self::MatchTag(pat) => config
                .unwrap()
                .tags()
                .iter()
                .any(|tag| pat.matches(tag.as_str())),
            Self::Union(v) => v.iter().any(|e| e.is_accepted(name, config)),
            Self::Inter(v) => v.iter().all(|e| e.is_accepted(name, config)),
            Self::Not(e) => !e.is_accepted(name, config),
            Self::Empty => false,
        }
    }

    /// Runs the filter against a set of NodeConfigs and returns the matched ones.
    pub fn filter_node_configs<'a, I>(&self, nodes: I) -> HashSet<NodeName>
    where
        I: Iterator<Item = (&'a NodeName, &'a NodeConfig)>,
    {
        if self == &Self::Empty {
            return HashSet::new();
        }

        nodes
            .filter_map(|(name, node)| {
                if self.is_accepted(name, Some(node)) {
                    Some(name)
                } else {
                    None
                }
            })
            .cloned()
            .collect()
    }

    /// Runs the filter against a set of node names and returns the matched ones.
    pub fn filter_node_names(&self, nodes: &[NodeName]) -> ColmenaResult<HashSet<NodeName>> {
        if self.has_node_config_rules() {
            Err(ColmenaError::Unknown {
                message: format!(
                    "Not enough information to run rule {:?} - We only have node names",
                    self
                ),
            })
        } else {
            Ok(nodes
                .iter()
                .filter_map(|name| {
                    if self.is_accepted(name, None) {
                        Some(name.clone())
                    } else {
                        None
                    }
                })
                .collect())
        }
    }
}

#[cfg(test)]
mod tests {
    use super::*;

    use std::collections::{HashMap, HashSet};

    macro_rules! node {
        ($n:expr) => {
            NodeName::new($n.to_string()).unwrap()
        };
    }

    #[test]
    fn test_empty_filter() {
        let filter = NodeFilter::new("").unwrap();
        assert_eq!(NodeFilter::Empty, filter);

        let filter = NodeFilter::new("\t").unwrap();
        assert_eq!(NodeFilter::Empty, filter);

        let filter = NodeFilter::new("    ").unwrap();
        assert_eq!(NodeFilter::Empty, filter);
    }

    #[test]
    fn test_empty_filter_rule() {
        assert!(NodeFilter::new(",").is_err());
        assert!(NodeFilter::new("a,,b").is_err());
        assert!(NodeFilter::new("a,b,c,").is_err());
    }

    #[test]
    fn test_filter_rule_mixed() {
        let filter = NodeFilter::new("@router,gamma-*").unwrap();
        assert_eq!(
            NodeFilter::Union(vec![
                Box::new(NodeFilter::MatchTag(GlobPattern::new("router").unwrap())),
                Box::new(NodeFilter::MatchName(GlobPattern::new("gamma-*").unwrap())),
            ]),
            filter,
        );

        let filter = NodeFilter::new("a, \t@b ,    c-*").unwrap();
        assert_eq!(
            NodeFilter::Union(vec![
                Box::new(NodeFilter::MatchName(GlobPattern::new("a").unwrap())),
                Box::new(NodeFilter::MatchTag(GlobPattern::new("b").unwrap())),
                Box::new(NodeFilter::MatchName(GlobPattern::new("c-*").unwrap())),
            ]),
            filter,
        );

        let filter = NodeFilter::new("a  & \t@b ,    c-*").unwrap();
        assert_eq!(
            NodeFilter::Inter(vec![
                Box::new(NodeFilter::MatchName(GlobPattern::new("a").unwrap())),
                Box::new(NodeFilter::Union(vec![
                    Box::new(NodeFilter::MatchTag(GlobPattern::new("b").unwrap())),
                    Box::new(NodeFilter::MatchName(GlobPattern::new("c-*").unwrap())),
                ])),
            ]),
            filter,
        );

        let filter = NodeFilter::new("( a  & \t@b ) ,  c-*").unwrap();
        assert_eq!(
            NodeFilter::Union(vec![
                Box::new(NodeFilter::Inter(vec![
                    Box::new(NodeFilter::MatchName(GlobPattern::new("a").unwrap())),
                    Box::new(NodeFilter::MatchTag(GlobPattern::new("b").unwrap())),
                ])),
                Box::new(NodeFilter::MatchName(GlobPattern::new("c-*").unwrap())),
            ]),
            filter,
        );

        let filter = NodeFilter::new("( a  & \t@b ) ,  !  c-*").unwrap();
        assert_eq!(
            NodeFilter::Union(vec![
                Box::new(NodeFilter::Inter(vec![
                    Box::new(NodeFilter::MatchName(GlobPattern::new("a").unwrap())),
                    Box::new(NodeFilter::MatchTag(GlobPattern::new("b").unwrap())),
                ])),
                Box::new(NodeFilter::Not(Box::new(NodeFilter::MatchName(
                    GlobPattern::new("c-*").unwrap()
                )))),
            ]),
            filter,
        );

        let filter = NodeFilter::new("( a  & \t@b ) ,  !!!  c-*").unwrap();
        assert_eq!(
            NodeFilter::Union(vec![
                Box::new(NodeFilter::Inter(vec![
                    Box::new(NodeFilter::MatchName(GlobPattern::new("a").unwrap())),
                    Box::new(NodeFilter::MatchTag(GlobPattern::new("b").unwrap())),
                ])),
                Box::new(NodeFilter::Not(Box::new(NodeFilter::MatchName(
                    GlobPattern::new("c-*").unwrap()
                )))),
            ]),
            filter,
        );
    }

    #[test]
    fn test_filter_node_names() {
        let nodes = vec![node!("lax-alpha"), node!("lax-beta"), node!("sfo-gamma")];

        assert_eq!(
            &HashSet::from_iter([node!("lax-alpha")]),
            &NodeFilter::new("lax-alpha")
                .unwrap()
                .filter_node_names(&nodes)
                .unwrap(),
        );

        assert_eq!(
            &HashSet::from_iter([node!("lax-alpha"), node!("lax-beta")]),
            &NodeFilter::new("lax-*")
                .unwrap()
                .filter_node_names(&nodes)
                .unwrap(),
        );
    }

    #[test]
    fn test_filter_node_configs() {
        // TODO: Better way to mock
        let template = NodeConfig {
            tags: vec![],
            target_host: None,
            target_user: None,
            target_port: None,
            allow_local_deployment: false,
            build_on_target: false,
            replace_unknown_profiles: false,
            privilege_escalation_command: vec![],
            extra_ssh_options: vec![],
            keys: HashMap::new(),
            system_type: None,
        };

        let mut nodes = HashMap::new();

        nodes.insert(
            node!("alpha"),
            NodeConfig {
                tags: vec!["web".to_string(), "infra-lax".to_string()],
                ..template.clone()
            },
        );

        nodes.insert(
            node!("beta"),
            NodeConfig {
                tags: vec!["router".to_string(), "infra-sfo".to_string()],
                ..template.clone()
            },
        );

        nodes.insert(
            node!("gamma-a"),
            NodeConfig {
                tags: vec!["controller".to_string()],
                ..template.clone()
            },
        );

        nodes.insert(
            node!("gamma-b"),
            NodeConfig {
                tags: vec!["ewaste".to_string()],
                ..template
            },
        );

        assert_eq!(4, nodes.len());

        assert_eq!(
            &HashSet::from_iter([node!("alpha")]),
            &NodeFilter::new("@web")
                .unwrap()
                .filter_node_configs(nodes.iter()),
        );

        assert_eq!(
            &HashSet::from_iter([node!("alpha"), node!("beta")]),
            &NodeFilter::new("@infra-*")
                .unwrap()
                .filter_node_configs(nodes.iter()),
        );

        assert_eq!(
            &HashSet::from_iter([node!("beta"), node!("gamma-a")]),
            &NodeFilter::new("@router,@controller")
                .unwrap()
                .filter_node_configs(nodes.iter()),
        );

        assert_eq!(
            &HashSet::from_iter([node!("beta"), node!("gamma-a"), node!("gamma-b")]),
            &NodeFilter::new("@router,gamma-*")
                .unwrap()
                .filter_node_configs(nodes.iter()),
        );

        assert_eq!(
            &HashSet::from_iter([]),
            &NodeFilter::new("@router&@controller")
                .unwrap()
                .filter_node_configs(nodes.iter()),
        );

        assert_eq!(
            &HashSet::from_iter([node!("beta")]),
            &NodeFilter::new("@router&@infra-*")
                .unwrap()
                .filter_node_configs(nodes.iter()),
        );

        assert_eq!(
            &HashSet::from_iter([node!("alpha")]),
            &NodeFilter::new("!@router&@infra-*")
                .unwrap()
                .filter_node_configs(nodes.iter()),
        );
    }
}
